/*************************************************************
 * delaunay2d.c:
 * 
 * 2-dimensional delaunay triangulation
 * Y Azuma, T Horinouchi
 * 
 * public functions:
 *  Delaunay2D
 *  free_triangles
 *  print_triangles
 *************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <delaunay.h>


struct Side{
  int id[2];
  struct Side *next;
};
typedef struct Side side;

static
triangle *alloc_triangle(){
  triangle *t;
  t = (triangle *)malloc(sizeof(triangle));
  t->next = NULL;
  return(t);
}


static
side *alloc_side(){
  side *s;
  s = (side *)malloc(sizeof(side));
  s->next = NULL;
  return(s);
}


static
triangle *init_triangle(int id0, int id1, int id2, double *px, double *py){
  double p0x, p1x, p2x, p0y, p1y, p2y, a1, a2, b1, b2, c1, c2;
  triangle *t;
  p0x = px[id0];
  p0y = py[id0];
  p1x = px[id1];
  p1y = py[id1];
  p2x = px[id2];
  p2y = py[id2];
  a1 = 2.0*(p1x - p0x);
  a2 = 2.0*(p2x - p0x);
  b1 = 2.0*(p1y - p0y);
  b2 = 2.0*(p2y - p0y);
  c1 = p0x*p0x - p1x*p1x + p0y*p0y -p1y*p1y;
  c2 = p0x*p0x - p2x*p2x + p0y*p0y -p2y*p2y;
  t = alloc_triangle();
  t->id[0] = id0;
  t->id[1] = id1;
  t->id[2] = id2;
  t->x = (b1*c2 - b2*c1)/(a1*b2 - a2*b1);
  t->y = (c1*a2 - c2*a1)/(a1*b2 - a2*b1);
  t->r2 = (p0x - t->x)*(p0x - t->x) + (p0y - t->y)*(p0y - t->y);
  t->next = NULL;
  return(t);
}


void free_triangle(triangle *t){
  triangle *w, *wz;
  w = t;
  while(w->next != NULL){
    wz = w->next; 
    free(w);
    w = wz; 
  }
}


static
void free_side(side *s){
  side *w, *wz;
  w = s;
  while(w->next != NULL){
    wz = w->next; 
    free(w);
    w = wz; 
  }
}

void print_triangles(char *title, triangle *t){
  triangle *w;
  w = t;
  printf("%s: ",title);
  if (t != NULL){
    do {
      printf("%d %d %d   ", w->id[0], w->id[1], w->id[2]);
      w = w->next;
    } while (w != NULL);
  }
  printf("\n");
}


static
void print_sides(char *title, side *s){
  side *v;
  v = s;
  printf("%s: ",title);
  if (s != NULL){
    do {
      printf("%d %d   ", v->id[0], v->id[1]);
      v = v->next;
    } while (v != NULL);
  }
  printf("\n");
}


static
void push_triangle(triangle **tp, triangle *x){
  triangle *w;
  if(*tp == NULL){
    *tp = x;
  } else {
    w = *tp;
    while(w->next != NULL){
      w = w->next;
    }
    w->next = x;
  }
}


/* ϡƤޤ໰ѷĺϡ*/
static
void set_outer_triangle_points(int n, double *px, double *py){ /*OK*/
  int i;
  double max_x, max_y, min_x, min_y, xwidth, ywidth;
  
  max_x = *px;
  for(i=0;i<n;i++){
    if(max_x < px[i]) max_x = px[i];
  }
  
  max_y = *py;
  for(i=0;i<n;i++){
    if(max_y < py[i]) max_y = py[i];
  }
  
  min_x = *px;
  for(i=0;i<n;i++){
    if(min_x > px[i]) min_x = px[i];
  }
  
  min_y = *py;
  for(i=0;i<n;i++){
    if(min_y > py[i]) min_y = py[i];
  }
  
  xwidth = max_x - min_x;
  ywidth = max_y - min_y;
  px[n] = min_x - xwidth;                         
  py[n] = min_y - ywidth;
  px[n+1] = max_x + 3.0*xwidth;   
  py[n+1] = min_y - ywidth;
  px[n+2] = min_x - xwidth;                       
  py[n+2] = min_y + 3.0*ywidth;
}


static
int triangle_has_point(triangle *t, double px, double py){
  if( (px - t->x)*(px - t->x) + (py - t->y)*(py - t->y) < t->r2 ) {
    return 1;
  } else {
    return 0;
    }
}


/*  ѷνTprκɸϡprܱ˴ޤ໰ѷ(t_pr_in)ϡ*/
static
triangle *triangle_pr_in(triangle **tp, double px, double py){
  triangle *t_pr_in, *w, *w0, *t;
  t = *tp;
  t_pr_in = NULL;
  w0 = NULL;
  w = t;
  do {
    if(triangle_has_point(w, px, py)){
      if(w0 == NULL){
	t = w->next;
      } else {
	w0->next = w->next;
      }
      w->next = NULL;
      if(t_pr_in == NULL){
	t_pr_in = w;
      } else {
	push_triangle(&t_pr_in, w);
      }
      if(w0 == NULL){
	w = t;
      } else {
	w = w0->next;
      }
    } else {
      if(w0 == NULL) {
	w0 = w;
      } else { 
	w0 = w0->next;
      }
      w = w->next;
    }
  } while(w != NULL);
  *tp = t;	
  return(t_pr_in);
}


static
side *triangle_to_sides(triangle *t){
  int x, y, z;
  side *s, *s1, *s2, *s3;
  s1 = alloc_side();
  s2 = alloc_side();
  s3 = alloc_side();
  x = t->id[0];
  y = t->id[1];
  z = t->id[2];
  s1->id[0] = (x < y)? x : y;
  s1->id[1] = (x < y)? y : x;
  s2->id[0] = (x < z)? x : z;
  s2->id[1] = (x < z)? z : x;
  s3->id[0] = (y < z)? y : z;
  s3->id[1] = (y < z)? z : y;
  s1->next = s2;
  s2->next = s3;
  s = s1;
  return(s);
}


/*  t_pr_inϡt_pr_in礷¿ѷγ¦(hull sides)ϡ*/   /*change*/
static
side *hull_sides(triangle *t_pr_in){
  side *S, *v, *v0, *ts, *tv, *tw;
  triangle *w;
  w = t_pr_in;
  /*  t_pr_inκǽλѷˤĤơ*/
  S =  triangle_to_sides(w);
  /*  t_pr_inΣܰʹߤλѷˤĤơ*/
  while(w->next != NULL){
    w = w->next;
    ts = triangle_to_sides(w);
    tv = ts;
    do {
      v = S;
      v0 = NULL;
      while(1){
	if(tv->id[0] == v->id[0] && tv->id[1] == v->id[1]){
	  if(v0 == NULL){
	    S = v->next;
	    free(v);
	    v = S;
	  }else{
	    v0->next = v->next;
	    free(v);
	    v = v0->next;
	  }
	  tw = tv;
	  tv = tv->next;
	  free(tw);
	  break;
	}else{
	  if(v->next == NULL){
	    tw = tv; 
	    tv = tv->next;
	    tw->next = NULL;
	    v->next = tw;
	    break;
	  }else{
	    v0 = v;
	    v = v->next;
	  }
	}
      } 
    } while(tv != NULL);		
  }
  free_triangle(t_pr_in);
  return(S);
}


/* hull sidesϡprhull sidesˤäƤǤ뻰ѷϡ*/   /*change*/
static
triangle *new_triangle(side *S, int i, double *px, double *py){
  triangle *t_pr, *w;
  side *v;
  t_pr = NULL;
  t_pr = init_triangle(S->id[0], S->id[1], i, px, py);
  v = S;
  while(v->next != NULL){
    v = v->next;
    w = init_triangle(v->id[0], v->id[1], i, px, py);
    if(t_pr == NULL){t_pr = w;}
    else{push_triangle(&t_pr, w);}
  }
  free_side(S);
  return(t_pr);
}


/* ¦ޤ໰ѷ */
static
void excluding_outer_triangles(int n, triangle **T){
  triangle *w, *w0;
  w = *T;
  w0 = NULL;
  do{
    if(w->id[0] >= n || w->id[1] >= n || w->id[2] >= n){
      if(w0 == NULL){
	*T = w->next;
	free(w);
	w = *T;
      }else{
	w0->next = w->next;
	free(w);
	w = w0->next;
      }
    }else{
      w0 = w;
      w = w->next;
    }
  }while(w != NULL);
}


/*  ϡDelaunayʬλѷνTϡ*/   /*change*/
triangle *Delaunay2D(int n, double *px, double *py){
  int i,j;
  triangle *T, *t_pr_in, *t_pr;
  side *S;
  set_outer_triangle_points(n, px, py);
  T = init_triangle(n, n+1, n+2, px, py); /* outer triangle */
  for(i=0;i<n;i++){
    t_pr_in = triangle_pr_in(&T, px[i], py[i]);
    S = hull_sides(t_pr_in);
    t_pr = new_triangle(S, i, px, py);
    push_triangle(&T, t_pr);
  }
  excluding_outer_triangles(n, &T);
  return(T);
}


/*  κɸե˽  */
static
void fprint_pxpy(int n, double *px, double *py){
  int i;
  FILE *fp_p;
  fp_p = fopen("pxpy.dat","w");
  fprintf(fp_p, "id     px        py\n");
  for(i=0;i<n;i++){
    fprintf(fp_p,"%d  %f  %f\n", i, px[i], py[i]);
  }
  fclose(fp_p);
}


/*  ѷνTե˽  */
static
void fprint_triangles(char *filename, triangle *T){
  triangle *w;
  FILE *fp_T;
  fp_T = fopen(filename,"w");
  fprintf(fp_T, "ids of triangles\n");
  w = T;
  do{
    fprintf(fp_T, "%d %d %d\n", w->id[0], w->id[1], w->id[2]);
    w = w->next;
  }while(w != NULL);
  fclose(fp_T);
}


/*  main  */
#ifdef TESTMAIN
int main(){
  
  int n=9;
  double *px, *py;
  triangle *T, *t_pr_in, *t_pr, *w;
  px = (double *)malloc((n+3)*sizeof(double));
  py = (double *)malloc((n+3)*sizeof(double));
  
  px[0] = 0;  py[0] = 0;
  px[1] = 4;  py[1] = 2;
  px[2] = 8;  py[2] = 0;
  px[3] = 10;  py[3] = 2;
  px[4] = 7;  py[4] = 5;
  px[5] = 2;  py[5] = 5;
  px[6] = 0;  py[6] = 10;
  px[7] = 5;  py[7] = 8;
  px[8] = 9;  py[8] = 9;

  fprint_pxpy(n, px, py);

  /*  set_outer_triangle_points(n, px, py);
  printf("set_outer_triangle_points\n");
  printf("n   %f %f\n",   px[n],   py[n]);
  printf("n+1 %f %f\n",   px[n+1], py[n+1]);
  printf("n+2 %f %f\n\n", px[n+2], py[n+2]);
	
  /*
    printf("init_triangle pure test (you can remove it)\n");
    T = init_triangle(0, 5, 6, px, py);
    w = init_triangle(1, 2, 3, px, py);
    printf("%d %d %d %f %f %f\n\n", T->id[0], T->id[1], T->id[2], T->x, T->y, T->r2);  
    push_triangle(&T, w);
    print_triangles("T",T);
    
    printf("triangle_has_point: %d %d\n",
    triangle_has_point(T,0.2,0.2),
    triangle_has_point(T,-0.2,-0.2));
    free_triangle(T);
    
    printf("push_triangle\n");
    T = NULL;
    push_triangle(&T, w);
    print_triangles("w",T);  */
  
/*
  printf("init_triangle (honban)\n");
  T = init_triangle(n, n+1, n+2, px, py);
  printf("%d %d %d %f %f %f\n\n", T->id[0], T->id[1], T->id[2], T->x, T->y, T->r2);  
	
  free_triangle(T);*/
  
  T = Delaunay2D(n, px, py);
  print_triangles("Delaunay2D",T);

  fprint_triangles("Delaunay2D.dat",T);

  free_triangle(T);
  return(0);
}
#endif
